#include "BuildSession.h"

#include "Logger.h"
#include "Directory.h"

BuildSession::BuildSession(const FileSystemPath& source_dir, 
    const FileSystemPath& build_dir)
    : m_build_environment(source_dir)
{
    auto working_build_dir = build_dir;
    if (working_build_dir.is_empty().value())
    {
        working_build_dir = FileSystemPath::current_dir().value();
    }
    m_build_environment.set_build_dir(working_build_dir / "build");
}

Status BuildSession::configure()
{
    LOG_INFO("Configuring project at:" << m_build_environment.get_source_dir());
    Vector<FileSystemPath> yaml_files;
    STATUS_CHECK(Directory::getFilesWithExtension(
        m_build_environment.get_source_dir(), 
        ".yaml", 
        yaml_files,
        true), "Error looking for build files");
    for(const auto& yaml_file : yaml_files)
    {
        if (yaml_file.file_name() == "build")
        {
            STATUS_CHECK(add_target(yaml_file),
                "Error adding target at: " + yaml_file.stem().str());
        }
    }
    return configure_tests();
}

Status BuildSession::add_target(const FileSystemPath& config_path)
{
    LOG_INFO("Adding target at: " << config_path);
    auto target = BuildTarget::create(config_path);
    STATUS_CHECK(target->configure(m_build_environment), "Error configuring target");
    m_targets.push_back(std::move(target));
    return {};
}

Status BuildSession::configure_tests()
{
    const auto test_dir = m_build_environment.get_test_dir();
    LOG_INFO("Scanning test files at:" << test_dir);
    Vector<FileSystemPath> candidate_test_files;
    STATUS_CHECK(Directory::getFilesWithExtension(
        test_dir, 
        ".cpp", 
        candidate_test_files,
        true), "Error looking for test files");

    Vector<FileSystemPath> test_files;
    for(const auto& test_file : candidate_test_files)
    {
        if (test_file.stem().starts_with("Test"))
        {
            if (test_file.contains("core"))
            {
                LOG_INFO("Adding " << test_file);
                test_files.push_back(test_file);
            }
        }
    }
    return {};
}

Status BuildSession::build()
{
    LOG_INFO("Starting build at: " << m_build_environment.get_build_dir());
    Directory::create(m_build_environment.get_build_dir(), true);
    for(auto& target : m_targets)
    {
        auto run_status = target->build();
        if (!run_status.ok())
        {
            return run_status;
        } 
    }
    return {};
}
