#include "BuildTarget.h"

#include "Logger.h"
#include "Directory.h"
#include "YamlParser.h"

#include "BuildLibrary.h"
#include "BuildExecutable.h"

Ptr<BuildTarget> BuildTarget::create(const FileSystemPath& build_config_path)
{
    auto build_config = Ptr<BuildConfig>::create(build_config_path);
    build_config->load();
    auto build_type = build_config->get_target_type();
    if (build_type == BuildTargetType::SHARED_LIBRARY || build_type == BuildTargetType::SHARED_LIBRARY)
    {
        Ptr<BuildTarget> target_type;
        target_type = Ptr<BuildLibrary>::create(std::move(build_config));
        return std::move(target_type);
    }
    else 
    {
        Ptr<BuildTarget> target_type;
        target_type = Ptr<BuildExecutable>::create(std::move(build_config));
        return std::move(target_type);
    }
}

BuildTarget::BuildTarget(BuildTargetType target_type)
    : m_type(target_type)
{
}

BuildTarget::BuildTarget(Ptr<BuildConfig> config)
    : m_build_config(std::move(config)),
      m_type(m_build_config->get_target_type())
{
}

void BuildTarget::add_dependency(BuildTarget* dependency)
{
    m_dependencies.push_back(dependency);
}

BuildTargetType BuildTarget::get_type() const 
{
    return m_type;
}

const Vector<BuildTarget*>& BuildTarget::get_dependencies() const
{
    return m_dependencies;
}

const Vector<String>& BuildTarget::get_dependency_names() const 
{
    return m_dependency_names;
}

const String& BuildTarget::get_name() const 
{
    return m_name;
}

void BuildTarget::get_public_include_dirs(Vector<FileSystemPath>& dirs) const 
{
    dirs.extend(m_include_dirs);
    for(auto target : m_dependencies)
    {
        target->get_public_include_dirs(dirs);
    }
}

Status BuildTarget::load_config()
{
    STATUS_CHECK(m_build_config->load(), "Failed to load build config.");
    m_name = m_build_config->get_directory_name();
    return {};
}

FileSystemPath BuildTarget::get_directory() const
{
    return m_build_config->get_directory();
}

Status BuildTarget::populate_include_dirs()
{
    const auto search_dir = get_directory();
    m_include_dirs.push_back(search_dir);
    return Directory::getAllSubDirectories(search_dir, m_include_dirs);
}

Status BuildTarget::configure(const BuildEnvironment& env)
{
    for(auto target : m_dependencies)
    {
        target->configure(env);
    }

    const auto search_dir = get_directory();
    LOG_INFO("Scanning build file at: " << search_dir);

    STATUS_CHECK(populate_include_dirs(), "Error populating include dirs");
    return {};
}