#pragma once

#include "Error.h"
#include "Pointer.h"
#include "Vector.h"
#include "Optional.h"

#include <functional>

template<typename T>
class List
{
public:
	List(const T& item)
		: m_item(item)
	{
	}

	List(T&& item)
		: m_item(std::move(item))
	{
	}

	List* next() 
	{
		return m_next.get();
	}

	const List* next() const 
	{
		return m_next.get();
	}

	const List* end() const 
	{
		if (is_end())
		{
			return this;
		}
		auto next_entry = next();
		while(next_entry)
		{
			if(!next_entry->is_end())
			{
				next_entry = next_entry->next();
			}
			else
			{
				break;
			}
		}
		return next_entry;
	}

	using onItemFunc = std::function<bool(const T& item)>;
	void forEach(onItemFunc item_func) const
	{
		if (is_end())
		{
			return;
		}
		auto next_entry = next();
		while(next_entry)
		{
			if (!item_func(next_entry->item()))
			{
				return;
			}
			next_entry = next_entry->next();
		}
	}

	List* end() 
	{
		if (is_end())
		{
			return this;
		}
		auto next_entry = next();
		while(next_entry)
		{
			if(!next_entry->is_end())
			{
				next_entry = next_entry->next();
			}
			else
			{
				break;
			}
		}
		return next_entry;
	}

	const List* find(const T& item) const 
	{
		if (m_item == item)
		{
			return this;
		}
		auto next_entry = next();
		while(next_entry)
		{
			if (next_entry->m_item == item)
			{
				return next_entry;
			}
			next_entry = next_entry->next();
		}
		return nullptr;
	}

	Optional<size_t> find_index(const T& item) const
	{
		if (item == m_item)
		{
			return {0};
		}
		auto next_entry = next();
		size_t count = 1;
		while(next_entry)
		{
			if (next_entry->m_item == item)
			{
				return {count};
			}
			next_entry = next_entry->next();
			count++;
		}
		return {};	
	}

	size_t size() const 
	{
		auto next_entry = next();
		size_t count = 1;
		while(next_entry)
		{
			next_entry = next_entry->next();
			count++;
		}
		return count;
	}

	void insert(const T& item)
	{
		auto end_entry = end();
		end_entry->set_next(Ptr<List>::create(item));
	}

	void insert(T&& item)
	{
		auto end_entry = end();
		end_entry->set_next(Ptr<List>::create(std::move(item)));
	}

	const T& item() const 
	{
		return m_item;
	}

	T& item() 
	{
		return m_item;
	}

	bool empty() const 
	{
		return is_end();
	}

	bool is_end() const 
	{
		return !bool(m_next);
	}

	void flatten(Vector<T>& items) const
	{
		items.push_back(m_item);
		auto next_entry = next();
		while(next_entry)
		{
			items.push_back(next_entry->item());
			next_entry = next_entry->next();
		}	
	}

private:
	void set_next(Ptr<List> entry)
	{
		m_next = std::move(entry);
	}

	T m_item;
	Ptr<List> m_next;
};
