#include "Directory.h"
#include "ConsoleLogger.h"

#include <dirent.h>
#include <errno.h>
#include <sys/stat.h>

#include <stdio.h>

Status Directory::getDirectoryContents(const FileSystemPath& path, 
    Vector<FileSystemPath>& ret, bool directoryOnly)
{
    Status status;
    errno = 0;
    auto dirp = ::opendir(path.str().raw());
    if (dirp == nullptr)
    {
        ON_ERRNO("Failed to get directory content in opendir");
    }

    while(auto ep = ::readdir(dirp))
    {
        auto rel_path = String(ep->d_name);
        if (rel_path == "." || rel_path == "..")
        {
            continue;
        }

        const auto full_path = path.join(rel_path);
        if (directoryOnly)
        {
            IF_OK_AND_TRUE(full_path.is_directory())
            {
                ret.push_back(full_path);
            }
        }
        else
        {
            IF_OK_AND_TRUE(full_path.is_regular_file_or_directory())
            {
                ret.push_back(full_path);
            }          
        }
    }
    String errno_msg;
    if (errno != 0)
    {
        errno_msg = "Failed traversing dirp structure | ";
        errno_msg += Error::from_errno();
    }
    errno = 0;
    const auto rc = ::closedir(dirp);
    if (rc != 0)
    {
        errno_msg += "Failed to close dirp";
        errno_msg += Error::from_errno();
    }
    if (!errno_msg.empty())
    {
        return Status(errno_msg);
    }
    return {};
}

Status Directory::getAllSubDirectories(const FileSystemPath& path, 
        Vector<FileSystemPath>& ret)
{
    const auto is_dir = path.is_directory();
    if (!is_dir.ok())
    {
        return Status(is_dir.error());
    }
    if (!is_dir.value())
    {
        return {};
    } 

    Vector<FileSystemPath> dir_contents;
    STATUS_CHECK(getDirectoryContents(path, dir_contents, true),
        "Failed to get directory contents");

    for (const auto& entry : dir_contents)
    {
        IF_OK_AND_TRUE(entry.is_directory())
        {
            ret.push_back(entry);
            STATUS_CHECK(getAllSubDirectories(entry, ret), 
                "Failed to get subdirectories");
        }
    }
    return {};
}

Status Directory::getFiles(const FileSystemPath& path, 
    Vector<FileSystemPath>& ret, 
    bool recursive,
    const String& extension)
{
    const auto is_dir = path.is_directory();
    if (!is_dir.ok())
    {
        return Status(is_dir.error());
    }
    if (!is_dir.value())
    {
        return {};
    }

    Vector<FileSystemPath> dir_contents;
    STATUS_CHECK(getDirectoryContents(path, dir_contents),
        "Failed to get directory contents");

    for (const auto& entry : dir_contents)
    {
        IF_OK_AND_TRUE(entry.is_regular_file())
        {
            if (!extension.empty())
            {
                if (entry.extension() == extension)
                {
                    ret.push_back(entry);
                }
            }
            else
            {
                ret.push_back(entry);
            }
        }
        else if(recursive)
        {
            IF_OK_AND_TRUE(entry.is_directory())
            {
                Vector<FileSystemPath> child_paths;
                STATUS_CHECK(getFiles(entry, child_paths, recursive, extension), "Failed to get files");
                ret.extend(child_paths);
            }
        }
    }
    return {};
}

Status Directory::getFilesWithExtension(const FileSystemPath& path, 
    const String& extension, 
    Vector<FileSystemPath>& ret, 
    bool recursive)
{
    return getFiles(path, ret, recursive, extension);
}

Status Directory::create(const FileSystemPath& path, bool existsOk)
{
    FileSystemPath working_path;
    if (path.is_directory().value())
    {
        working_path = path;
    }
    else
    {
        working_path = path.parent_path();
    }

    if (working_path.exists())
    {
        if (!existsOk)
        {
            return Status(Error(
                "Attempted to create existing directory."
                ));
        }
        return {};
    }
    else
    {
        errno = 0;
        mode_t mode{S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH};
        Status status;
        const auto rc = ::mkdir(working_path.str().raw(), 
            mode);
        if (rc !=0)
        {
            status.on_errno("Failed to create directory.");
        }
        return status;
    }
    return {};
}
