#include "File.h"

#include "Char.h"
#include "Directory.h"
#include "Result.h"
#include "FilePosixImpl.h"

File::File(const FileSystemPath& path)
    : m_impl(Ptr<FilePosixImpl>::create()),
      m_path(path)
{
}

File::~File()
{
    close();
}

Status File::close()
{
    return m_impl->do_close();
}

bool File::exists() const
{
    return m_path.exists();
}

String File::getExtension() const
{
    return m_path.extension();
}

FileFormat::Format File::inferFormat() const
{
    const auto extension = getExtension();
    return FileFormat::inferFormat(extension);
}

Status File::open(AccessMode accessMode)
{
    if (m_path.is_absolute() && !m_path.parent_path().exists())
    {
        Directory::create(m_path.parent_path(), true);
    }
    return m_impl->do_open(m_path, accessMode);
}

Status File::read(VecBytes& buffer)
{
    if (const auto rc = open(AccessMode::Read); !rc.ok())
    {
        return rc;
    }

    if (const auto rc = m_impl->update_size(); !rc.ok())
    {
        return rc;
    }

    buffer.resize(m_impl->m_size);
    const auto result = m_impl->do_read(buffer);
    if (!result.ok())
    {
        return Status(result.error());
    }
    buffer.resize(result.value());
    return {};
}

Status File::read(String& ret)
{
    if (const auto rc = open(AccessMode::Read); !rc.ok())
    {
        return rc;
    }

    VecBytes buffer;
    if (const auto rc = m_impl->update_size(); rc.ok() && m_impl->m_size > 0)
    {
        buffer.resize(m_impl->m_size);
        const auto result = m_impl->do_read(buffer);
        if (!result.ok())
        {
            return Status(result.error());
        }
        buffer.resize(result.value());
        ret.append(buffer);
    }
    else
    {
        buffer.resize(1024);
        while(true)
        {
            const auto result = m_impl->do_read(buffer);
            if (!result.ok())
            {
                return Status(result.error());
            }
            if (result.value() < 1024)
            {
                if (result.value() > 0)
                {
                    buffer.resize(result.value());
                    ret.append(buffer);
                }
                break;
            }
            else
            {
                ret.append(buffer);
            }
        }
    }
    return {};
}

Status File::write(const String& text)
{
    bool had_to_open{false};
    Status status;
    if (!m_impl->is_open_for_write())
    {
        had_to_open = true;
        status = m_impl->do_open(m_path, File::AccessMode::Write);
        if (!status.ok())
        {
            return status;
        }
    }
    const auto result = m_impl->do_write(text.data(), text.data().size() - 1);
    if (!result.ok())
    {
        return Status(result.error());
    }

    if (had_to_open)
    {
        status = m_impl->do_close();
    }
    return status;
}