#include "FileSystemPath.h"

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>

FileSystemPath::FileSystemPath()
{
}

FileSystemPath::FileSystemPath(const String& path)
    : m_path(path)
{
}

FileSystemPath::FileSystemPath(const char* path)
    : m_path(path)
{

}

const String& FileSystemPath::str() const
{
    return m_path;
}

Result<bool> FileSystemPath::is_empty() const
{
    ERROR_IF_NOT_OK_OR_TRUE(is_regular_file(),
        "Requested empty check but target is not regular file.");
    const auto size_result = get_size();
    if (!size_result.ok())
    {
        return {size_result.error().msg()};
    }
    return {size_result.value() == 0};
}

Result<size_t> FileSystemPath::get_size() const
{
    ERROR_IF_NOT_OK_OR_TRUE(is_regular_file(),
        "Requested size check but target is not regular file.");

    struct stat buf;
    errno = 0;
    const auto rc = ::stat(m_path.raw(), &buf);
    if (rc != 0)
    {
        ON_ERRNO_RESULT("Failed to stat file");
    }
    return buf.st_size;
}

Result<FileSystemPath> FileSystemPath::current_dir()
{
    errno = 0;
    const auto path_max = ::pathconf(".", _PC_PATH_MAX);
    size_t size{0};
    if (path_max == -1)
    {
        size = 1024;
    }
    else if(path_max > 10240)
    {
        size = 10240;
    }
    else 
    {
        size = path_max;
    }

    Vector<char> buffer(path_max);
    errno = 0;
    const auto ret = ::getcwd(buffer.data(), path_max);
    if (ret == nullptr)
    {
        ON_ERRNO_RESULT("Failed to get cwd");
    }
    return FileSystemPath(String(buffer.data()));
}

Result<bool> FileSystemPath::is_regular_file() const 
{
    struct stat path_stat;
    const auto rc = ::stat(m_path.raw(), &path_stat);
    if (rc != 0)
    {
        ON_ERRNO_RESULT("Failed to stat file");
    } 
    return S_ISREG(path_stat.st_mode);
}

Result<bool> FileSystemPath::is_directory() const
{
    struct stat path_stat;
    const auto rc = ::stat(m_path.raw(), &path_stat);
    if (rc != 0)
    {
        ON_ERRNO_RESULT("Failed to stat file");
    } 
    return S_ISDIR(path_stat.st_mode);
}

Result<bool> FileSystemPath::is_regular_file_or_directory() const
{
    struct stat path_stat;
    const auto rc = ::stat(m_path.raw(), &path_stat);
    if (rc != 0)
    {
        ON_ERRNO_RESULT("Failed to stat file");
    } 
    return S_ISREG(path_stat.st_mode) || S_ISDIR(path_stat.st_mode);
}

bool FileSystemPath::is_absolute() const 
{
    return false;
}

FileSystemPath FileSystemPath::parent_path() const
{
    const auto split = m_path.rsplit('/');
    return FileSystemPath(split.first());
}

String FileSystemPath::extension() const 
{
    const auto split = m_path.rsplit('.');
    String result(".");
    result += split.second();
    return result;
}

FileSystemPath FileSystemPath::stem() const
{
    const auto split = m_path.rsplit('/');
    if (split.second().empty())
    {
        return split.first();
    }
    else
    {
        return split.second();
    }
}

bool FileSystemPath::starts_with(const String& prefix) const
{
    return m_path.starts_with(prefix);
}

bool FileSystemPath::contains(const String& term) const
{
    return m_path.contains(term);
}

String FileSystemPath::file_name() const
{
    String name_and_ext;
    const auto split = m_path.rsplit('/');
    if (split.second().empty())
    {
        name_and_ext = split.first();
    }
    else
    {
        name_and_ext = split.second();
    }
    return name_and_ext.rsplit('.').first();
}

bool FileSystemPath::exists() const 
{
    return ::access(m_path.raw(), F_OK) == 0;
}

FileSystemPath FileSystemPath::join(const String& entry) const
{
    auto new_path = *this;
    new_path.m_path += m_delimiter;
    new_path.m_path += entry;
    return new_path;
}